/*
 * ipblocks -- string parsing conversion utility for parsing subnet/bitmask notation
 *
 * gcc -Wall -O3 -c ipblocks.c -o ipblocks.o && ar rc ipblocks.a ipblocks.o && ranlib ipblocks.a
 *
 * skape
 * 06/01/2002
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include <arpa/inet.h>

#include <string.h>

#include "ipblocks.h"

#define CHAR_VAL(ptr, elem) ((unsigned char *)ptr)[elem]
#define CHAR_ADDR(ptr, elem) ((unsigned char *)ptr)+elem

/*
 * accepted strings:
 *
 * 1.2.3.4
 * 10.0.0.0/24
 * 10/24
 * 10.0.0.0/255.255.255.0
 * 10/255.255.255.0
 *
 */

void _ipblockScanStringIp(const char *buf, unsigned long *out)
{
	short a, b, c, d;
	a = b = c = d = 0;

	sscanf(buf, "%hi.%hi.%hi.%hi", &a, &b, &c, &d);

	CHAR_VAL(out,0) = a;
	CHAR_VAL(out,1) = b;
	CHAR_VAL(out,2) = c;
	CHAR_VAL(out,3) = d;
}

void _ipblocksIncrementIp(unsigned long *ip)
{
	if (CHAR_VAL(ip,3) == 255)
	{
		if (CHAR_VAL(ip,2) == 255)
		{
			if (CHAR_VAL(ip,1) == 255)
			{
				if (CHAR_VAL(ip,0) == 255)
					return;
		
				CHAR_VAL(ip,0)++;
				CHAR_VAL(ip,1) = 0;	
			}
			else
			{
				CHAR_VAL(ip,1)++;
				CHAR_VAL(ip,2) = 0;
			}
		}
		else
		{
			CHAR_VAL(ip,2)++;
			CHAR_VAL(ip,3) = 0;
		}
	}
	else
		CHAR_VAL(ip,3)++;
}

short ipblocksParseString(const char *blockString, unsigned long *subnet, unsigned long *netmask, unsigned char *bitmask)
{
	char *ptr, subnetBuffer[64], netmaskBuffer[64];
	int pos;
	unsigned long tempMask = 0;

	*subnet = *netmask = *bitmask = 0;

	if (!blockString)
		return 0;

	/* If no bitmask/netmask supplied, assume /32 */
	if (!(ptr = strchr(blockString, '/')))
	{
		*subnet  = inet_addr(blockString);
		*netmask = 0xFFFFFFFF;
		*bitmask = 32;

		return 0;
	}

	pos = ptr - blockString;		

	memset(subnetBuffer, 0, sizeof(subnetBuffer));
	memset(netmaskBuffer, 0, sizeof(netmaskBuffer));

	strncpy(subnetBuffer, blockString, pos & (sizeof(subnetBuffer)-1));
	strncpy(netmaskBuffer, blockString + (pos+1), sizeof(netmaskBuffer) - 1);

	_ipblockScanStringIp(subnetBuffer,subnet);
	_ipblockScanStringIp(netmaskBuffer,netmask);
	
	if ((CHAR_VAL(netmask,0) != 0) &&
		 (CHAR_VAL(netmask,1) == 0))
	{
		*bitmask = CHAR_VAL(netmask, 0);
		tempMask = (0xFFFFFFFF << (32 - *bitmask));

		*netmask = ntohl(tempMask);			
	}
	else
		for (pos = 0; pos < 31 && *netmask & (1 << pos); pos++, *bitmask = *bitmask + 1);

	*subnet = *subnet & *netmask;
	
	return 1;
}

unsigned long ipblocksGetIpCountString(const char *blockString)
{
	unsigned long subnet, netmask;
	unsigned char bitmask;

	if (ipblocksParseString(blockString, &subnet, &netmask, &bitmask) == 0)
		return 0;

	return (1 << (32 - bitmask));
}

IPBLOCK *ipblocksCreateIpBlock(const char *blockString, unsigned long entriesPerSlot, unsigned short flags)
{
	IPBLOCK *newBlock = (IPBLOCK *)malloc(sizeof(IPBLOCK));
	unsigned long ipCount, ipCountCurr, currentIp;
	int x = 0, y;

	if ((!newBlock) || (!ipblocksParseString(blockString, &newBlock->subnet, &newBlock->netmask, &newBlock->bitmask)))
		return newBlock;

	ipCount = newBlock->ipCount = (1 << (32 - newBlock->bitmask));
	currentIp = newBlock->currentIpContext = newBlock->subnet;

	newBlock->currentIpCount = 0;

	newBlock->tableElements = ((ipCount % entriesPerSlot)?1:0) + (ipCount / entriesPerSlot);

	if (!(flags & IPBLOCK_TABLEIZE))
		return newBlock;
	
	newBlock->table = (IPBLOCK_TABLE *)malloc(sizeof(IPBLOCK_TABLE) * newBlock->tableElements);

	if (!newBlock->table)
		return newBlock;

	for (ipCountCurr = ipCount; x < newBlock->tableElements; x++, ipCountCurr -= entriesPerSlot)
	{
		if (ipCountCurr < entriesPerSlot)
			newBlock->table[x].entries = ipCountCurr;
		else
			newBlock->table[x].entries = entriesPerSlot;

		newBlock->table[x].ips = (unsigned long *)malloc(sizeof(unsigned long) * newBlock->table[x].entries);

		for (y = 0; y < newBlock->table[x].entries; y++)
		{
			newBlock->table[x].ips[y] = currentIp;

			_ipblocksIncrementIp(&currentIp);
		}
	}

	return newBlock; 
}

unsigned long ipblocksGetIpCount(IPBLOCK *ipblock)
{
	return (ipblock)?ipblock->ipCount:0;
}

int ipblocksEnumerate(IPBLOCK *ipblock, unsigned long *array, unsigned long elements)
{
	int x = 0;

	memset(array, 0, sizeof(unsigned long) * elements);

	while ((ipblock->currentIpCount < ipblock->ipCount) && (x < elements))
	{
		array[x] = ipblock->currentIpContext;

		_ipblocksIncrementIp(&ipblock->currentIpContext);

		x++; ipblock->currentIpCount++;
	}		

	return x;
}

void ipblocksReset(IPBLOCK *ipblock)
{
	ipblock->currentIpCount = 0;
	ipblock->currentIpContext = ipblock->subnet;
}

void ipblocksDestroyIpBlock(IPBLOCK *ipblock)
{
	if (!ipblock)
		return;

	if (ipblock->table)
	{
		int x = 0;

		for (; x < ipblock->tableElements; x++)
		{
			if (ipblock->table[x].ips)
				free(ipblock->table[x].ips);
		}

		free(ipblock->table);
	}

	free(ipblock);
}

#ifdef TEST

int main(int argc, char **argv)
{
	IPBLOCK *blockTableized = ipblocksCreateIpBlock("10.0.254.0/24",25,IPBLOCK_TABLEIZE);
	IPBLOCK *blockNonTableized = ipblocksCreateIpBlock("10.0.0.0/24",25,0);
	unsigned long ips[80];
	int y, x, ret;

	if (!blockTableized)
		return 0;

	for (x = 0; x < blockTableized->tableElements; x++)
	{
		for (y = 0; y < blockTableized->table[x].entries; y++)
		{
			fprintf(stdout,"%hi.%hi.%hi.%hi\n",
							CHAR_VAL(&blockTableized->table[x].ips[y],0),
							CHAR_VAL(&blockTableized->table[x].ips[y],1),
							CHAR_VAL(&blockTableized->table[x].ips[y],2),
							CHAR_VAL(&blockTableized->table[x].ips[y],3));
		}
	}

	ipblocksDestroyIpBlock(blockTableized);

	while ((ret = ipblocksEnumerate(blockNonTableized, ips, 80)))
	{
		for (x = 0; x < ret; x++)
			fprintf(stdout,"%hi.%hi.%hi.%hi\n",
							CHAR_VAL(&ips[x],0),
							CHAR_VAL(&ips[x],1),
							CHAR_VAL(&ips[x],2),
							CHAR_VAL(&ips[x],3));

	}

	ipblocksDestroyIpBlock(blockNonTableized);

	return 1;
}

#endif
